/* ====================================================================
 * FTD4Linux Software License, Version 1.0
 *
 * Copyright (c) 2000-2004 The FTD4Linux Development Team.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        FTD4Linux Development Team (http://www.ftd4linux.2y.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "FTD" and "FTD4Linux" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact alguno@ftd4linux.2y.net.
 *
 * 5. Products derived from this software may not be called "FTD4Linux",
 *    nor may "FTD4Linux" appear in their name, without prior written
 *    permission of the FTD4Linux Development Team.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE FTD4LINUX DEVELOPMENT TEAM OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include "ftd4linux.h"
#include "ftd_app_core.h"
#include "ftd_category.h"
#include "ftd_events_forum.h"

/**
 * Leest een extra gegevens van een event en het forum en de bijbehorende reacties uit een XML gecodeerde string.
 *
 * @arg     buffer      De XML gecodeerde tekst waar de gegevens in staan.
 * @arg     event       Het event waarin de gegevens opgeslagen moeten worden.
 *
 * @return  0 indien het uitlezen geod ging.
 */
int
ftd_events_forum_parse(FtdString *buffer, FtdEvent *event)
{
    xmlDocPtr doc;
    xmlNodePtr cur;

    event->forum = ftd_events_forum_new();

    g_assert(core);
    g_assert(buffer);

    doc = xmlParseMemory(ftd_string_value(buffer), ftd_string_length(buffer));
    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr,"ftd_events_forum_parse(): empty document!\n");
        xmlFreeDoc(doc);
        return 1;
    }

    if (strcasecmp(cur->name, "Forum")) {
        fprintf(stderr,"ftd_events_forum_parse(): document of the wrong type, root node != Forum");
        xmlFreeDoc(doc);
        return 2;
    }

    cur = cur->xmlChildrenNode;
    if (!cur) {
        g_warning("ftd_events_forum_parse(): no children of root-node!\n");
        xmlFreeDoc(doc);
        return 3;
    }

    while (cur) {
        if (!strcasecmp(cur->name, "Event")) {
            xmlChar *code;
            int numcode;

            code = xmlGetProp(cur, "Code");
            if (code == NULL) {
                continue;
            }

            numcode = atoi(code);
            xmlFree(code);

            /* In deze routine moeten we alleen berichten met <Event Code="1"> opslaan */
            if (numcode == 1 || numcode == 2) {    // code 1 = nieuw bericht; code 2 = edited bericht
                ftd_events_forum_parse_forum(doc, cur->xmlChildrenNode, event, (numcode == 2) ? TRUE : FALSE);
            }
            if (numcode == 0 || numcode == 4) {     // code 0 = stamp; code 4 = rest van de data
                ftd_events_forum_parse_data(doc, cur->xmlChildrenNode, event);
            }

            // Event code 3 == kudos
            if (numcode == 3) {
                ftd_events_forum_parse_kudos(doc, cur->xmlChildrenNode, event);
            }
        }
        cur = cur->next;
    }
    xmlFreeDoc(doc);

    return 0;
}

/**
 * Leest gegevens van een event uit XML-boom.
 *
 * @arg     doc         Het XML document met daarin de gegevens.
 * @arg     cur         Een pointer naar het element in het XML document waar de gegevens staan.
 * @arg     event       Het event waarin de gegevens opgeslagen dienen te worden.
 */
void
ftd_events_forum_parse_data(xmlDocPtr doc, xmlNodePtr cur, FtdEvent *event)
{
    while(cur != NULL) {
        if (!strcasecmp(cur->name, "Kudos")) {
            event->kudos = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "Pars")) {
            event->aantalParFiles = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "ParType")) {
            event->parType = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "BlockSize")) {
            event->parBlockSize = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "Status")) {
            event->status = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "View")) {
            event->numViews = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "Block")) {
            // Geen flauw idee wat dit inhoudt... kwam het toevallig tegen
        }
        if (!strcasecmp(cur->name, "Stamp")) {
            event->forum->lastUpdated = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "Locked")) {
            event->locked = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "Reason")) {
            event->reason = ftd_xml_parse_string(doc, cur);
        }

        cur = cur->next;
    }
}

/**
 * Leest een forum reactie uit een XML boom.
 *
 * @arg     doc         Het XML document met daarin de forum reactie.
 * @arg     cur         Een pointer naar het element in het XML document waar de forum reactie staat.
 * @arg     event       Het event waarin het forum opgeslagen moet worden.
 * @arg     edited      True indien de te lezen reactie een bewerkt bericht is.
 */
void
ftd_events_forum_parse_forum(xmlDocPtr doc, xmlNodePtr cur, FtdEvent *event, gboolean edited)
{
    FtdEventForumReactie *reactie;

    reactie = ftd_events_forum_reactie_new();

    // Verwerk de XML data
    while(cur != NULL) {
        if (!strcasecmp(cur->name, "MessID")) {
            reactie->messageId = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "Author")) {
            reactie->author = ftd_xml_parse_string(doc, cur);
        }
        if (!strcasecmp(cur->name, "Body")) {
            reactie->body = ftd_xml_parse_string(doc, cur);
        }
        if (!strcasecmp(cur->name, "Stamp")) {
            reactie->stamp = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "Edit")) {
            reactie->lastEdited = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "Reply")) {
            reactie->reply = ftd_xml_parse_int(doc, cur);
        }
        if (!strcasecmp(cur->name, "Avatar")) {
            reactie->avatar = ftd_xml_parse_string(doc, cur);
        }
        cur = cur->next;
    }

    /* Is het bericht verwijderd ? */
    if (edited && reactie->lastEdited < 0) {
        ftd_events_forum_reactie_free(reactie);
        return;
    }

    event->forum->forum_reacties = ftd_list_append(event->forum->forum_reacties, reactie);
}

/**
 * Leest een kudo-donatie uit een XML boom.
 *
 * @arg     doc         Het XML document waarin de kudo-donatie staat.
 * @arg     cur         Een pointer naar het element in het XML document waar de kudo-donatie staat.
 * @arg     event       Het event waarin de kudo-donatie opgeslagen moet worden.
 */
void
ftd_events_forum_parse_kudos(xmlDocPtr doc, xmlNodePtr cur, FtdEvent *event)
{
    FtdString *user = NULL;
    FtdList *list;
    int kudos = 0;

    // Verwerk de XML data
    while(cur != NULL) {
        if (!strcasecmp(cur->name, "User")) {
            user = ftd_xml_parse_string(doc, cur);
        }
        if (!strcasecmp(cur->name, "Kudos")) {
            kudos = ftd_xml_parse_int(doc, cur);
        }
        cur = cur->next;
    }

    // Zoek alle bijbehorende forumberichten en verwerk het kudos aantal
    list = event->forum->forum_reacties;
    while (list) {
        FtdEventForumReactie *reactie = ftd_list_value(list);

        if (!strcmp(ftd_string_value(reactie->author), ftd_string_value(user))) {
            reactie->kudos = kudos;
        }

        list = ftd_list_next(list);
    }
}
